package org.kaveh.commons.farsi.utils;

import org.joda.time.DateTime;
import org.joda.time.chrono.GregorianChronology;
import org.joda.time.chrono.JulianChronology;
import org.kaveh.commons.farsi.access.JalaliDate;
import org.kaveh.commons.farsi.date.JalaliDateImpl;

/**
 * 
 * Utilities to support JalaliDate class.
 * 
 * @author Kaveh Ranjbar
 * 
 */
public class JalaliCalendarUtils {

	private static final double PERSIAN_EPOCH = 1948320.5;

	public static JalaliDate convertGregorianToJalali(DateTime gregorianDate) {
		return convertJulianDateNumberToJalali(convertJulianDateToJulianDateNumber(new DateTime(
				gregorianDate, JulianChronology.getInstance())));
	}

	public static DateTime convertJalaliToGregorian(JalaliDate jalaliDate) {
		return new DateTime(
				convertJulianDateNumberToJulianDate(convertJalaliToJulianDateNumber(jalaliDate)),
				GregorianChronology.getInstance());
	}

	public static int getWeekDay(JalaliDate jalaliDate) {
		return convertJalaliToGregorian(jalaliDate).getDayOfWeek();
	}

	private static double convertJalaliToJulianDateNumber(JalaliDate jd) {
		double epbase, epyear;
		double year = jd.getYear();
		double month = jd.getMonth();
		double day = jd.getDay();

		epbase = year - ((year >= 0) ? 474 : 473);
		epyear = 474 + (epbase % 2820);

		return day
				+ ((month <= 7) ? ((month - 1) * 31) : (((month - 1) * 30) + 6))
				+ Math.floor(((epyear * 682) - 110) / 2816) + (epyear - 1)
				* 365 + Math.floor(epbase / 2820) * 1029983
				+ (PERSIAN_EPOCH - 1);
	}

	private static JalaliDate convertJulianDateNumberToJalali(double jd) {
		double year, month, day, depoch, cycle, cyear, ycycle, aux1, aux2, yday;

		jd = Math.floor(jd) + 0.5;

		depoch = jd
				- convertJalaliToJulianDateNumber(new JalaliDateImpl(475, 1, 1));
		cycle = Math.floor(depoch / 1029983);
		cyear = (depoch % 1029983);
		if (cyear == 1029982) {
			ycycle = 2820;
		} else {
			aux1 = Math.floor(cyear / 366);
			aux2 = (cyear % 366);
			ycycle = Math
					.floor(((2134 * aux1) + (2816 * aux2) + 2815) / 1028522)
					+ aux1 + 1;
		}
		year = ycycle + (2820 * cycle) + 474;
		if (year <= 0) {
			year--;
		}
		yday = (jd - convertJalaliToJulianDateNumber(new JalaliDateImpl((int) year,
				1, 1))) + 1;
		month = (yday <= 186) ? Math.ceil(yday / 31) : Math
				.ceil((yday - 6) / 30);
		day = (jd - convertJalaliToJulianDateNumber(new JalaliDateImpl((int) year,
				(int) month, 1))) + 1;
		return new JalaliDateImpl((int) year, (int) month, (int) day);
	}

	private static double convertJulianDateToJulianDateNumber(
			DateTime julianDate) {
		if (!julianDate.getChronology().equals(JulianChronology.getInstance())) {
			throw (new RuntimeException("Julian date is required"));
		}
		int year = julianDate.getYear();
		int month = julianDate.getMonthOfYear();
		int day = julianDate.getDayOfMonth();
		/* Adjust negative common era years to the zero-based notation we use. */

		if (year < 1) {
			year++;
		}

		/*
		 * Algorithm as given in Meeus, Astronomical Algorithms, Chapter 7, page
		 * 61
		 */

		if (month <= 2) {
			year--;
			month += 12;
		}

		return ((Math.floor((365.25 * (year + 4716)))
				+ Math.floor((30.6001 * (month + 1))) + day) - 1524.5);
	}

	private static DateTime convertJulianDateNumberToJulianDate(double jd) {
		double z, a, b, c, d, e, year, month, day;

		jd = jd + 0.5;
		z = Math.floor(jd);

		a = z;
		b = a + 1524;
		c = Math.floor((b - 122.1) / 365.25);
		d = Math.floor(365.25 * c);
		e = Math.floor((b - d) / 30.6001);

		month = Math.floor((e < 14) ? (e - 1) : (e - 13));
		year = Math.floor((month > 2) ? (c - 4716) : (c - 4715));
		day = b - d - Math.floor(30.6001 * e);

		/*
		 * If year is less than 1, subtract one to convert from a zero based
		 * date system to the common era system in which the year -1 (1 B.C.E)
		 * is followed by year 1 (1 C.E.).
		 */

		if (year < 1) {
			year--;
		}

		return new DateTime((int) year, (int) month, (int) day, 0, 0, 0, 0,
				JulianChronology.getInstance());
	}

}
